import gradio as gr
from bleak import BleakScanner
import asyncio
from bleak import BleakClient
import struct
from yams.bluetooth_device import characteristic_bat, get_device_status
from apscheduler.schedulers.background import BackgroundScheduler


device_info = {}
dev_status = {}

async def bleak_scan(filter_key):
    global device_info
    devices = await BleakScanner.discover()
    for d in devices:
        # print(dir(d))
        # print(d.name, d.address)
        name = f"{d.name}"
        addr = f"{d.address}"

        if filter_key in name:
            device_info[f"{addr} - {name}"] = addr

    print(device_info)


def search_bt_devices(filter_key):
    global device_info
    asyncio.run(bleak_scan(filter_key))

    return gr.CheckboxGroup(choices=list(device_info.keys()), 
                            value=list(device_info.keys()),
                            label="Available devices")


def get_dev_status(devices):
    info = {}
    # print(info)

    scheduler = BackgroundScheduler()
    scheduler.add_job(get_dev_status_job, "interval", seconds=10, args=[devices]) 
    scheduler.start()

    return gr.Button("update status", interactive=True)

def get_dev_status_job(devices):
    global dev_status
    info = {}

    try:
        for dev in devices:
            info[dev] = asyncio.run(get_device_status(device_info[dev]))
        print(info)
        dev_status = info
    except Exception as e:
        print(str(e))


async def erase_dev(addr):
    rst_char = "da39c934-1d81-48e2-9c68-d0ae4bbd351f"
    try:
        async with BleakClient(addr) as client:
            gr.Info(f"Erasing {client.address}")
            value = struct.pack("<I", int(68))
            await client.write_gatt_char(rst_char, value)
    except Exception as e:
        print(str(e))
        gr.Error(f"⚠️{str(e)}")

async def write_dev(addr, val, characteristics="da39c931-1d81-48e2-9c68-d0ae4bbd351f"):
    try:
        async with BleakClient(addr) as client:
            gr.Info(f"collection control {client.address} {val}")
            value = struct.pack("<I", int(val))
            await client.write_gatt_char(characteristics, value)
    except Exception as e:
        print(str(e))
        gr.Error(f"⚠️{str(e)}")


async def get_dev(addr, key):
    info = {addr: 'null'}
    try:
        async with BleakClient(addr) as client:
            value = await client.read_gatt_char(key)
            info[addr] = value
    except Exception as e:
        print(str(e))
        gr.Error(f"⚠️{str(e)}")

    return info


def erase_flash_data(available_devices):
    gr.Info("Erasing flash data...")

    print(available_devices)
    for k in available_devices:
        asyncio.run(erase_dev(device_info[k]))
        
    return gr.Number(value=None, label="Erase code"), gr.Checkbox(label="Enable erase feature", value=False), gr.Button("Erase flash data", interactive=False)


def update_status():
    global dev_status
    print(dev_status)
    return gr.JSON(dev_status, visible=True)


def bt_scanner_interface():
    text = gr.Text("MSense", label="Device filter", scale=2)

    bt_search = gr.Button("Search Bluetooth devices 📱")

    with gr.Row():
        available_devices = gr.CheckboxGroup(label="Available devices")
        with gr.Column():
            bt_status = gr.Button("(Experimental) Get device status ✅")
            bt_update_status = gr.Button("(Experimental) Update status", interactive=False)

    dev_panel = gr.JSON(visible=False)
    
    bt_search.click(search_bt_devices, inputs=[text], outputs=available_devices)    
    bt_status.click(get_dev_status, inputs=available_devices, outputs=bt_update_status)
    bt_update_status.click(update_status, outputs=dev_panel)
    

    with gr.Accordion(label="Device control", open=True):
        # conect_btn = gr.Button("Connect selected", interactive=False)
        # memo_page = gr.Text(label="Status memo")

        with gr.Row():
            start_btn = gr.Button("Start▶️")
            stop_btn = gr.Button("Stop🛑")

        start_btn.click(collection_ctl_start, inputs=[available_devices])
        stop_btn.click(collection_ctl_stop, inputs=[available_devices])
        

    # erase control
    with gr.Accordion(label="🚨🚨🚨Danger zone🚨🚨🚨", open=False):
        erase_passcode = gr.Number(label="Erase code")
        erase_enable = gr.Checkbox(label="Enable erase feature")
        
        erase_btn = gr.Button("Erase flash data", interactive=False)
        erase_btn.click(erase_flash_data, inputs=[available_devices],
                        outputs=[erase_passcode, erase_enable, erase_btn])

        erase_enable.change(set_erase_feature, inputs=[erase_enable, erase_passcode], outputs=[erase_btn])

def collection_ctl_start(devices):
    gr.Info("Collection control starting... 🦭")
    collection_ctl(devices, True)
    gr.Info("✅✅✅ All done ✅✅✅")

def collection_ctl_stop(devices):
    gr.Info("Collection control stoping... 🛑")
    collection_ctl(devices, False)
    gr.Info("✅✅✅ All done ✅✅✅")

def collection_ctl(devices, start_collect):
    if start_collect:
        val = 1
    else:
        val = 0

    for k in devices:
        asyncio.run(write_dev(device_info[k], val))

def set_erase_feature(erase_enable, erase_passcode):
    if erase_enable:
        if erase_passcode == 68:
            gr.Warning("Erase feature is enabled!")
            return gr.Button("Erase flash data", interactive=True)
        else:
            gr.Warning("Incorrect password")
            return gr.Button("Erase flash data", interactive=False)
    else:
        gr.Info("Erase feature disabled")
        return gr.Button("Erase flash data", interactive=False)





# original credit: https://github.com/SenSE-Lab-OSU/MotionSenseHRV4Flash/blob/main/DataExtraction/data_extraction.py
import os
import sys
import struct
import re
import datetime
import traceback
import argparse
import pandas as pd
import re
import numpy as np
import numpy
from datetime import datetime
import gradio as gr

def data_extraction_interface():
    # in_files = gr.File(file_count="multiple")
    in_dir = gr.Text("/path/to/binary/data", label="Input directory")
    out_dir = gr.Text("/path/to/output", label="Output directory")

    btn = gr.Button("Extract raw data")
    btn.click(main, inputs=[in_dir, out_dir])


def process_data_test(data) -> int:
    errors = 0
    check_array = [1, 2, 3, 4, 5, 6]
    for data_byte in range(0, len(data), 2):
        result = struct.unpack("<h", data[data_byte:data_byte + 2])[0]
        # print(result)
        inx = data_byte // 2 % 6
        if (result != check_array[inx]):
            # print("error: got " + str(result[0]) + " expected " + str(check_array[data_byte//2 % 6]))
            errors += 1
    print("errors: " + str(errors))
    return errors


def calculate_file_end(file):
    index = -1
    subtract_length = 0
    while file[index] == 0xff:
        subtract_length += 1
        index -= 1
    return subtract_length


struct_key = {"f": 4,
              "h": 2,
              "I": 4,
              "i": 4,
              "H": 2,
              "Q": 8
              }


def process_data(data, categories: list[list], format: list, use_check=False) -> int:
    # we always assume that the last category is the packet counter
    # assert len(categories) == len(format)
    errors = 0

    skip_code = 1
    counter_value = 1
    current_index = 0
    data_position = 0
    num_of_categories = len(categories)
    end_trim_size = calculate_file_end(data)
    data_length = len(data) - end_trim_size
    while data_position + struct_key[format[current_index][1]] <= data_length:
        try:

            length = struct_key[format[current_index][1]]
            data_byte = data[data_position:data_position + length]
            data_position += length
            result = struct.unpack(format[current_index], data_byte)[0]

            if result == 4294967295:
                continue
            # print(result)
            categories[current_index].append(result)
            current_index += 1
            if current_index >= num_of_categories:
                current_index = 0
                if use_check:
                    # we are on the last category, which means we are looking at the counter
                    if (result != counter_value):
                        print("error: got " + str(result) + " expected " + str(counter_value))
                        errors += 1
                    counter_value += 1


        except Exception as e:
            errors += 1
            print(e)
    
    if len(categories[0]) > len(categories[1]):
        print("0xff-trim issue found, fixing...")
        categories[0].pop()

    resultant_end_trim_workaround(categories)

    if len(categories[0]) > len(categories[1]):
        print("0xff-trim issue found, fixing...")
        categories[0].pop()

    resultant_end_trim_workaround(categories)

    print("errors: " + str(errors))
    return errors


def resultant_end_trim_workaround(categories):
    length_array = []
    for element in categories:
        length_array.append(len(element))

    max_diff = numpy.max(length_array) - numpy.min(length_array)
    if max_diff == 0:
        return
    if max_diff == 1:
        print("warning: end length of array differs by 1. Implementing fix.")
        print("mismatch array: " + str(length_array))
        max_value = max(length_array)
        for element in categories:
            if max_value == len(element):
                element.pop()
    elif max_diff > 1:
        print("error: end lengths of array differs by too much. Data is potentially corrupt.")
        print("mismatch array: " + str(length_array))


    max_diff = numpy.max(length_array) - numpy.min(length_array)
    if max_diff == 0:
        return
    if max_diff == 1:
        print("warning: end length of array differs by 1. Implementing fix.")
        print("mismatch array: " + str(length_array))
        max_value = max(length_array)
        for element in categories:
            if max_value == len(element):
                element.pop()
    elif max_diff > 1:
        print("error: end lengths of array differs by too much. Data is potentially corrupt.")
        print("mismatch array: " + str(length_array))

def file_sort(element1: str):
    numeric_index = element1.find(it_prefix)
    numeric_time = element1[numeric_index + len(it_prefix):len(element1)]
    return int(re.sub("\D", "", numeric_time))


def gather_files_by_prefix(prefix: str, path):
    global it_prefix
    it_prefix = prefix
    all_files = []
    files = os.listdir(path)
    for file in files:
        if file.startswith(prefix) and file.endswith('.bin'):
            all_files.append(file)
    all_files.sort(key=file_sort)
    return all_files


def obtain_prefix_ids(path):
    all_files = []
    files = os.listdir(path)
    for file in files:
        if file[0].isdigit():
            id = re.search(r'\d+', file)
            if id is not None:
                id = id.group()
                if id not in all_files:
                    all_files.append(id)

    return all_files


def counter_validity_check(df: pd.DataFrame):
    counter_columns = df.iloc[:, -1:]
    counter_arr = numpy.array(counter_columns).flatten()
    diff_arr = numpy.diff(counter_arr)
    check_array = (diff_arr == 5) | (diff_arr == 10) | (diff_arr < -65000)
    print("pass counter check: " + str(numpy.all(check_array)))
    print("and number of non matching samples: " + str(numpy.count_nonzero(check_array == 0)))


def collect_all_data_by_prefix(path, prefix: str, labels: list[str], types: list[str]):
    total_errors = 0

    all_data = []
    for element in range(len(labels)):
        all_data.append([])

    files = gather_files_by_prefix(prefix, path)  

    if len(files) == 0:
        return
    for file in files:
        full_path = os.path.join(path, file)
        print("full path: ", full_path)
        test_file = open(full_path, "rb")
        data = test_file.read()
        if len(data) != 0:
            total_errors += process_data(data, all_data, types)
        else:
            print("Warning: found empty file!")

    full_dict = {}
    for index in range(len(labels)):
        full_dict[labels[index]] = all_data[index]

    dataset = pd.DataFrame(full_dict)
    dataset = get_cdct(dataset, files)

    return dataset


def unit_conversion_ac(data_set):
    for c in ['AccX', 'AccY', 'AccZ']:
        data_set[c] = data_set[c] /(2**16-1)*8
    return data_set


def generate_csv_for_pattern(in_dir, type_prefix: str, search_key: str, labels, formats, plot=False, out_dir="./"):
    file_name = f"{type_prefix}"
    print(type_prefix, search_key)
    data_set = collect_all_data_by_prefix(in_dir, search_key, labels, formats)
    if data_set is not None:
        os.makedirs(out_dir, exist_ok=True)
        counter_validity_check(data_set)

        if 'ac' in search_key:
            print("perform unit conversion for IMU")
            data_set = unit_conversion_ac(data_set)

        data_set.to_csv(os.path.join(out_dir, file_name))
        if plot: raise NotImplementedError
        # if plot: graph_generation.pd_graph_generation(search_key, data_set)

# get init time
def get_t0(file_list):
    pattern = r'\d*[A-Za-z]+(\d+)\.bin$'
    t = sorted([int(match.group(1)) for filename in file_list if (match := re.search(pattern, filename))])
    return t[0]

# CDCT: calculated data collection time
def get_cdct(df, bin_list, fs=320):
    t0 = get_t0(bin_list)

    # fs = 10*32 if 'ac' in bin_list[0] else 5*64
    counter_diff = np.diff(df['Counter']) % (2^16 - 1)
    counter_diff = np.insert(counter_diff, 0, 0)

    df['CDCT'] = t0 + np.cumsum(counter_diff) / fs
    df['Datetime'] = [datetime.utcfromtimestamp(int(t)) for t in df['CDCT']]
    return df

def main(in_dir, out_dir, gradio=True):
    ppg_labels = ["ir1", "ir2", "g1", "g2",  "Timestamp", "Counter"]
    ppg_formats = ["<i", "<i", "<i", "<i", "<i", "<i"]


    acc_labels = ["AccX", "AccY", "AccZ", "GyroX", "GyroY", "GyroZ", "ENMO", "Timestamp", "Counter"]
    acc_formats = ["<h", "<h", "<h", "<f", "<f", "<f", "<f", "<i", "<i"]

    try:
        ids = obtain_prefix_ids(in_dir)
        ids.append("")

        for id in ids:
            search_prefix = id + "ac"
            file_name = search_prefix + ".csv"
            generate_csv_for_pattern(in_dir, file_name, search_prefix, acc_labels, acc_formats, out_dir=out_dir)
            search_prefix = id + "ppg"
            file_name = search_prefix + ".csv"
            generate_csv_for_pattern(in_dir, file_name, search_prefix, ppg_labels, ppg_formats, out_dir=out_dir)

        if gradio: gr.Info("✅ Extraction completed")
        print("operation completed.")
    except Exception as e:
        # gr.Error(str(e)) if gradio else print(str(e))
        print(str(e))
        print("operation completed with error")


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--in_dir', type=str, help="directory where binary files are located")
    parser.add_argument('-o', '--out_dir', type=str, default="./", help="output directory")

    args = parser.parse_args()

    main(args.in_dir, args.out_dir, gradio=False)
    



import gradio as gr
import pandas as pd
import matplotlib.pyplot as plt

class DataExplorer():
    def __init__(self):
        self.df = None

    def interface(self):
        file = gr.File(file_types=['.csv'])

        with gr.Row():
            with gr.Column():
                x_lim_low = gr.Slider(label="x start")
                x_range = gr.Slider(label="x range", minimum=0, maximum=10000)

            with gr.Column():
                y_lim_low = gr.Slider(label='y_start', maximum=5e5)
                y_range = gr.Slider(label="y range", minimum=0, maximum=5e5)

        fig = gr.Plot()

        x_to_plot = gr.Radio(label="x data")
        y_to_plot = gr.CheckboxGroup(label="y data")

        btn = gr.Button()
        btn.click(self.update_figure, inputs=[x_to_plot, y_to_plot, x_lim_low, x_range, y_lim_low, y_range], outputs=fig)    

        file.change(self.process_file, inputs=file, outputs=[x_to_plot, y_to_plot, x_lim_low])

    def process_file(self, file):
        self.df = pd.read_csv(file)
        choices = list(self.df.columns)
        return [gr.Radio(choices, label="x data"), 
                gr.CheckboxGroup(choices, label="y data"), 
                gr.Slider(label="x start", maximum=len(self.df.index))]
    
    def update_figure(self, x_to_plot, y_to_plot, 
                      xlim, xrange,
                      ylim, yrange):
        print(y_to_plot)
        fig = plt.figure(figsize=[10, 4])
        for y in y_to_plot:
            plt.plot(self.df[x_to_plot], self.df[y], label=y)
        plt.legend()
        plt.xlim(xlim, xlim + xrange)
        plt.ylim(ylim, ylim + yrange)
        return fig


